package edu.uiuc.dsi;

import java.net.Socket;
import java.net.ServerSocket;
import java.io.IOException;
import java.util.ArrayList;
import java.lang.Thread;
import java.lang.Runnable;


public abstract class Core implements Runnable
{
	private ArrayList<TCPSocket> objTCPSocketList = new ArrayList<TCPSocket>();
	private ServerSocket objServerSocket;
	private ArrayList<Base> objComponentList = new ArrayList<Base>();
	private TCPSocket objSelfTCPSocket;
	
	
	public Core(int port, String selfSocketName) throws IOException
	{
		objServerSocket = new ServerSocket(port);
		objSelfTCPSocket = new TCPSocket(this, selfSocketName);
		
		Thread objThread = new Thread(this);
		objThread.start();
	}
	
	public void writeStatusMessage(Object sender, String message)
	{
		System.out.println("[" + sender.getClass().getName() + "] " + message);
	}
	
	@Override
	public void run()
	{
		System.out.println("Listening on " + objServerSocket.getLocalPort() + "...");

		
		Socket objSocket;
		
		try
		{
			while ((objSocket = objServerSocket.accept()) != null)
			{
				System.out.println("Socket accepted!");
	
				TCPSocket objTCPSocket = new TCPSocket(this, objSocket);
				objTCPSocketList.add(objTCPSocket);
	
				Thread objThread = new Thread(objTCPSocket);
				objThread.start();
				
				writePacket(objTCPSocket, new Packet(getSelfTCPSocket().getSocketName().getBytes(), Packet.PACKET_TYPE_CONNECTION_NAME));
				objTCPSocket.setConnectionReadyState(true);
			}
			
			if (!objServerSocket.isClosed())
				objServerSocket.close();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		
	}
	
	public TCPSocket connect(String hostname, int port) throws IOException
	{
		System.out.println("Attempting connection to " + hostname + ":" + port);

		java.net.Socket objSocket = new Socket(hostname, port);
		
		TCPSocket objTCPSocket = new TCPSocket(this, objSocket);
		objTCPSocketList.add(objTCPSocket);

		Thread objThread = new Thread(objTCPSocket);
		objThread.start();
		
		System.out.println("Connection made to " + hostname + ":" + port);
		
		return objTCPSocket; 
	}
	
	public TCPSocket getTCPSocket(int id)
	{
		return objTCPSocketList.get(id);
	}
	
	public void write(TCPSocket objTCPSocket, byte[] data)
	{
		Packet objPacket = new Packet(data);
		objPacket.setPacketType(Packet.PACKET_TYPE_USER_DATA);
		
		writePacket(objTCPSocket, objPacket);
	}
	
	public void writePacket(TCPSocket objTCPSocket, Packet objPacket)
	{
		if (objTCPSocket == this.getSelfTCPSocket())
		{
			System.out.println("[W/R] data=" + new String(objPacket.getFullPacketBytes()));
			//Thread objThread = new Thread(new CoreThread(this, objTCPSocket, objPacket));
			//objThread.start();
			deliverPacket(objTCPSocket, objPacket);
		}
		else
			objTCPSocket.writeToSocket(objPacket.getFullPacketBytes());
	}
	
	public void broadcast(byte[] data)
	{
		Packet objPacket = new Packet(data);
		objPacket.setPacketType(Packet.PACKET_TYPE_USER_DATA);
		
		broadcastPacket(objPacket);
	}
	
	public synchronized void broadcastPacket(Packet objPacket)
	{
		writePacket(this.getSelfTCPSocket(), objPacket);
		for (int i=0; i < objTCPSocketList.size(); i++)
			writePacket(objTCPSocketList.get(i), objPacket);
	}
	
	public void addComponent(Base objComponent)
	{
		objComponentList.add(objComponent);
		objComponent.setApplicationCore(this);
	}
	
	public synchronized void deliverPacket(TCPSocket objTCPSocket, Packet objPacket)
	{
		switch (objPacket.getPacketType())
		{
			case Packet.PACKET_TYPE_COMPONENT_DATA:
				for (Base objComponent : objComponentList)
				{
					if (objComponent.getComponentName().equals(objPacket.getTargetComponentName()))
					{
						objComponent.processPacket(objTCPSocket, objPacket);
						break;
					}
				}
				break;
				
			case Packet.PACKET_TYPE_CONNECTION_NAME:
				objTCPSocket.setSocketName(new String(objPacket.getData()));
				
				if (!objTCPSocket.isConnectionReady())
				{
					writePacket(objTCPSocket, new Packet(getSelfTCPSocket().getSocketName().getBytes(), Packet.PACKET_TYPE_CONNECTION_NAME));
					objTCPSocket.setConnectionReadyState(true);
				}
				
				break;
				
			case Packet.PACKET_TYPE_USER_DATA:
				processPacket(objTCPSocket, objPacket);
				break;
		}
	}
	
	public TCPSocket getSelfTCPSocket()
	{
		return objSelfTCPSocket;
	}
	
	public TCPSocket[] getActiveTCPSockets()
	{
		TCPSocket[] objReturnArray = new TCPSocket[objTCPSocketList.size()];
		
		for (int i=0; i < objTCPSocketList.size(); i++)
			objReturnArray[i] = objTCPSocketList.get(i);
		
		return objReturnArray;
	}
	
	public abstract void processPacket(TCPSocket sender, Packet packet);
	
	
}
