package edu.uiuc.dsi.apps;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import java.util.concurrent.Semaphore;

import java.nio.ByteBuffer;

import edu.uiuc.dsi.Packet;
import edu.uiuc.dsi.TCPSocket;
import edu.uiuc.dsi.mutex.MutExException;

public class DistributedObject extends edu.uiuc.dsi.Base
{
	private edu.uiuc.dsi.mutex.Base objMutExAlgorithm;
	private Serializable objDistributedObject = null;
	private Semaphore objSemaphoreLock = new Semaphore(0);
	
	private static final byte MESSAGE_REQUEST_OBJECT = 0x01;
	private static final byte MESSAGE_SENDING_OBJECT = 0x02;
	
	public DistributedObject(String componentName, edu.uiuc.dsi.mutex.Base objMutExAlgorithm)
	{
		super(componentName);
		this.objMutExAlgorithm = objMutExAlgorithm;
	}
	
	public void setObject(Serializable objDistributedObject)
	{
		this.objDistributedObject = objDistributedObject;
	}
	
	public Serializable requestObject() throws MutExException, InterruptedException
	{
		System.out.println("[edu.uiuc.dsi.apps.DistributedObject] MutEx Requested.");
		objMutExAlgorithm.requestMutEx();
		System.out.println("[edu.uiuc.dsi.apps.DistributedObject] MutEx Granted, Object Requested.");
		this.getObject();
		System.out.println("[edu.uiuc.dsi.apps.DistributedObject] Object Granted.");
		
		return objDistributedObject;
	}
	
	public void releaseObject() throws MutExException
	{
		objMutExAlgorithm.releaseMutEx();
	}
	
	private void getObject() throws InterruptedException
	{
		if (objDistributedObject == null)
		{
			broadcast( new byte[] { MESSAGE_REQUEST_OBJECT } );
			objSemaphoreLock.acquire();
		}
			
	}

	@Override
	public void processPacket(TCPSocket sender, Packet packet)
	{
		ByteBuffer objByteBuffer = ByteBuffer.wrap(packet.getData()), objResponseByteBuffer;
		byte packetType = objByteBuffer.get();
		
		switch (packetType)
		{
			case MESSAGE_REQUEST_OBJECT:
				if (objDistributedObject != null)
				{
					try
					{
						ByteArrayOutputStream objByteSerializationStream = new ByteArrayOutputStream();
						ObjectOutputStream objSerializationStream = new ObjectOutputStream(objByteSerializationStream);
						objSerializationStream.writeObject(objDistributedObject);
						
						objDistributedObject = null;
						
						objResponseByteBuffer = ByteBuffer.allocate(1 + objByteSerializationStream.toString().getBytes().length);
						objResponseByteBuffer.put(MESSAGE_SENDING_OBJECT);
						objResponseByteBuffer.put(objByteSerializationStream.toString().getBytes());						
						this.write(sender, objResponseByteBuffer.array());
						
						objSerializationStream.close();
						objByteSerializationStream.close();
					}
					catch (Exception e)
					{
						e.printStackTrace();
					}
				}
				
				break;
				
			case MESSAGE_SENDING_OBJECT:
				try
				{
					int iSeralizedClassSize = objByteBuffer.limit() - 1;
					byte[] bBuffer = new byte[iSeralizedClassSize];
					objByteBuffer.get(bBuffer, 0, iSeralizedClassSize);
					
					ByteArrayInputStream objByteSerializationStream = new ByteArrayInputStream(bBuffer);
					ObjectInputStream objSerializationStream = new ObjectInputStream(objByteSerializationStream);
					objDistributedObject = (Serializable)objSerializationStream.readObject();
					objSerializationStream.close();
					objByteSerializationStream.close();
					
					objSemaphoreLock.release();
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
				
				break;
				
		}
	}

	@Override
	public void initializeChildControls()
	{
		this.getApplicationCore().addComponent(objMutExAlgorithm);
	}

	
}
